# -*- coding: utf-8 -*-
" pyradio -- Console radio player. "

version_info = (0, 9, 3, 11, 31)

# Set it to True if new stations have been
# added to the package's stations.csv
stations_updated = False


__version__ = version = '.'.join(map(str, version_info))
__project__ = __name__
__author__ = "Ben Dowling"
__license__ = "MIT"
